#include <iostream>
#include <limits> //for infinity
#include <math.h>
#include <string>
#include <vector>
#include <cassert>

#include "cftree.h"

//
//------------------------------------------------------Node---------------------------------------------------------------
//

//Constructor
CFTree::Node::Node(bool isleaf, int dim, int branchingfactor): n(0), squaredsum(0), numchildren(0), isleaf(isleaf), parent(nullptr), dim(dim), branchingfactor(branchingfactor){
    linsum = new float_type[dim](); // initialized with all zeros
    centroid = new float_type[dim]();
    //set children to array of nullpointers, size of array = branchingfactor
    children = new Node*[branchingfactor]();
}

//Destructor
CFTree::Node::~Node(){
    delete[] linsum;
    delete[] centroid;
    //free each of the pointers in children before deallocating chilren itself
    for (int i = 0; i < branchingfactor; ++i){
        if (children[i]) delete children[i];
    }
    delete[] children;
}

void CFTree::Node::calculate_centroid(){
    if (n == 0) return; //can't calculate a centroid if there are no points in the cluster
    for (int i = 0; i < dim; ++i){
        centroid[i] = linsum[i] / n;
    }
}

const float_type *CFTree::Node::get_centroid(){
    return centroid;
}

bool CFTree::Node::append_child(Node *node){
    if (numchildren == branchingfactor) //no more room on this node
        return false;

    //set parent-child relationship pointers
    children[numchildren++] = node;
    node->parent = this;

    //update clustering features of all nodes on the path from the new node to the root
    Node *update = this;
    while(update){
        update->n += node->n;
        update->squaredsum += node->squaredsum;
        for (int i = 0; i < dim; ++i){
            update->linsum[i] += node->linsum[i];
        }
        update = update->parent;
    }//end while

    return true;
}

//this does NOT free node! (in a split node will still be needed, just getting a new parent)
bool CFTree::Node::remove_child(Node *node){
    for (int i = 0; i < numchildren; ++i){
        if (children[i] == node){

            //overwrite pointer that should be deleted with the last pointer in the array (order of children is irrelevant)
            children[i] = children[numchildren-1];
            children[numchildren-1] = nullptr;
            //no need to adjust i because we're skipping the rest of the loop anyways

            numchildren--;

            //update clustering features of all nodes on the path from the new node to the root
            Node *update = this;
            while(update){
                update->n -= node->n;
                update->squaredsum -= node->squaredsum;
                for (int i = 0; i < dim; ++i){
                    update->linsum[i] -= node->linsum[i];
                }
                update = update->parent;
            }//end while

            return true;
        }
    }//end for
    return false;
}



//prints out the Node, and calls this function on the Node's children
void CFTree::Node::print(std::string indent, bool last){
    //taken from: https://stackoverflow.com/questions/1649027/how-do-i-print-out-a-tree-structure

    std::cout << indent;
    if (last){
        std::cout << "\\-";
        indent += "    ";
    }
    else{
        std::cout << "|-";
        indent += "|   ";
    }
    std::cout << "[ N = " << n << ", Linear Sum = ( ";
    for (int i = 0; i < dim; ++i)
        std::cout << linsum[i] << " ";
    std::cout << "), Squared Sum = " << squaredsum;

    if (isleaf){ //assigned IDs are only stored in leaves
        std::cout << ", Assigned points: ( ";
        for (int id: assigned_points)
            std::cout << id << " ";
        std::cout << ")";
    }

    std::cout  << " ]" << std::endl;

    for (int i = 0; i < numchildren; ++i){
        children[i]->print(indent, i == numchildren - 1);
    }
}


//only prints out the current node (does not call the Node's children's print method)
void CFTree::Node::print_self(){
    std::cout << "[ N = " << n << ", Linear Sum = ( ";
    for (int i = 0; i < dim; ++i)
        std::cout << linsum[i] << " ";
    std::cout << "), Squared Sum = " << squaredsum;

    if (isleaf){ //assigned IDs are only stored in leaves
        std::cout << ", Assigned points: ( ";
        for (int id: assigned_points)
            std::cout << id << " ";
        std::cout << ")";
    }

    std::cout << " ]" << std::endl;
}




//
//----------------------------------------------------------------CFTree-----------------------------------------------------
//

//Constructor
CFTree::CFTree(float_type threshold, int wavelet_length, int used_resolution, int wavelets_per_point, int branchingfactor):
threshold(threshold), wavelet_length(wavelet_length), used_resolution(used_resolution), wavelets_per_point(wavelets_per_point),
branchingfactor(branchingfactor), num_points(0), root(new Node(true, used_resolution * wavelets_per_point, branchingfactor)){
    leaves.push_back(root); //initial root node is the first leaf node
}

//Destructor
CFTree::~CFTree(){
    delete root;
}




//---------------------------------------tree building methods--------------------------------------------------------------------------------

CFTree::Node *CFTree::find_leaf(const float_type *point){
    Node *current = root;

    float_type dist, current_dist;
    int closest_child;

    while(true){
        if (current->isleaf){
            #ifdef VERBOSE
            std::cout << "Found leaf" << std::endl;
            #endif

            return current;
        }
        else{
            dist = std::numeric_limits<float_type>::infinity();
            closest_child = -1;

            #ifdef VERBOSE
            std::cout << current->numchildren << " children" << std::endl;
            #endif

            for (int i = 0; i < current->numchildren; ++i){
                current->children[i]->calculate_centroid(); //centroid is not automatically updated, this call is needed to make sure centroid is up to date
                current_dist = dist_to_centroid(point, current->children[i]->get_centroid());

                #ifdef VERBOSE
                std::cout << "Distance to child " << i << ": " << current_dist << std::endl;
                #endif

                if (current_dist <= dist){
                    dist = current_dist;
                    closest_child = i;
                }
            }//end for

            assert(closest_child > -1);

            #ifdef VERBOSE
            std::cout << "Choosing child " << closest_child << " with distance " << dist << std::endl;
            #endif
            current = current->children[closest_child];
        }//end else
    }//end while(true)
}



void CFTree::modify_leaf(const float_type *point, Node *leaf, int point_id){

    float_type squaredsum = calculate_squaredsum(point);

    if (leaf->n == 0){ //no entries yet on that leaf (only happens when the first point is inserted into the initial root node)
        //simply insert clustering feature of the point into the leaf
        set_as_linsum(leaf, point);
        leaf->squaredsum = squaredsum;
        leaf->n++;
        leaf->assigned_points.push_back(point_id);

        #ifdef VERBOSE
        std::cout << "Inserting point into empty leaf node" << std::endl;
        #endif
    }
    else{
        float_type diameter = compute_diameter(leaf, point);
        if (diameter <= threshold){ //node can absorb new point

            #ifdef VERBOSE
            std::cout << "Adding point to existing cluster, combined diameter is " << diameter << std::endl;
            #endif

            //add the point to the CF of all nodes on the path from the leaf to the root
            leaf->assigned_points.push_back(point_id); //IDs are only stored in the actual leaf, not in its ancestors
            while(leaf){
                add_to_linsum(leaf, point);
                leaf->squaredsum += squaredsum;
                leaf->n++;

                leaf = leaf->parent;
            }
        }
        else{ //need to create new node
            #ifdef VERBOSE
            std::cout << "Creating new cluster, combined diameter " << diameter << " exceeds threshold" << std::endl;
            #endif

            Node *newleaf = new Node(true, leaf->dim, leaf->branchingfactor);
            set_as_linsum(newleaf, point);
            newleaf->squaredsum = squaredsum;
            newleaf->n++;
            newleaf->assigned_points.push_back(point_id);
            leaves.push_back(newleaf);

            insert_sibling_node(leaf, newleaf);

        }
    }//end else
}//end modify_leaf



void CFTree::insert_sibling_node(Node *node, Node *newnode){
    Node *parent=nullptr, *newparent=nullptr;

    while(true){
        if ( !(node->parent) ){ //root split
            newparent = new Node(false, node->dim, node->branchingfactor);
            newparent->append_child(node);
            newparent->append_child(newnode);
            root = newparent;

            #ifdef VERBOSE
            std::cout << "Root split" << std::endl;
            #endif

            break;
        }
        else if(node->parent->numchildren < branchingfactor){ //can simply append new node to tree
            node->parent->append_child(newnode);

            #ifdef VERBOSE
            std::cout << "Appending new leaf to existing parent" << std::endl;
            #endif

            break;
        }
        else{ //parent exists, but has no room for another child => split somewhere along the tree

            #ifdef VERBOSE
            std::cout << "Split - creating a new parent node" << std::endl;
            #endif

            parent = node->parent;
            newparent = new Node(false, node->dim, node->branchingfactor);

            //find the two nodes that are farthest apart from among all children of parent and newnode
            Node *a=nullptr, *b=nullptr;
            find_farthest_pair(parent, newnode, &a, &b);
            //now a and b point to the two nodes that are farthest apart

            //redistribute nodes so that a and b go to different parents, and all other nodes go to the same parent as whichever of a and b they are closest to
            redistribute_nodes(parent, newparent, newnode, a, b);


            //in the next iteration, newparent has to be inserted as sibling of parent
            node = parent;
            newnode = newparent;
            parent = nullptr;
            newparent = nullptr;

        }//end case split-somewhere-along-tree
    }//end while
}




//find the farthest pair of nodes from among all children of parent plus newnode
//result: *a and *b will point to the two nodes that are farthest apart
void CFTree::find_farthest_pair(Node *parent, Node *newnode, Node **a, Node **b){

    int limit = parent->numchildren;


    //making sure the centroids are up to date
    for (int i = 0; i < limit; ++i){
        parent->children[i]->calculate_centroid();
    }
    newnode->calculate_centroid();


    Node *child;
    float_type dist = -1;
    float_type current_dist;

    for (int i = 0; i < limit; ++i){ //for each child of parent:
        child = parent->children[i];

        //calculate distance to all subsequent children
        for (int j = i+1; j < limit; ++j){ //this will be skipped for the final value of i
            current_dist = dist_between_nodes(child, parent->children[j]);
            if (current_dist > dist){
                dist = current_dist;
                *a = child;
                *b = parent->children[j];
            }
        }//end j

        //calculate distance to newnode
        current_dist = dist_between_nodes(child, newnode);
        if (current_dist > dist){
            dist = current_dist;
            *a = child;
            *b = newnode;
        }
    }//end i
}




void CFTree::redistribute_nodes(Node *parent, Node *newparent, Node *newnode, Node *a, Node *b){
    #ifdef VERBOSE
    std::cout << "Redistributing " << parent->numchildren << " children (plus one newly created node)" << std::endl;
    #endif

    float_type dist_to_a, dist_to_b;

    //redistribute the nodes that are currently children of parent
    Node *child;
    for (int i = 0; i < parent->numchildren; ++i){ //reevaluation of parent->numchildren in each iteration is necessary
        child = parent->children[i];
        dist_to_a = dist_between_nodes(child, a);
        dist_to_b = dist_between_nodes(child, b);

        if (dist_to_b < dist_to_a){ //node has to be appended to newparent
            newparent->append_child(child);
            parent->remove_child(child);
            #ifdef VERBOSE
            std::cout << "Appending child " << i << " to new parent with distance " << dist_to_b << " to b (versus " << dist_to_a << " to a)" << std::endl;
            #endif

            //remove_child: current position gets overwritten with a new pointer => need to adjust index
            i--;
            #ifdef VERBOSE
            std::cout << "Adjusting index" << std::endl;
            #endif

        }
        else { //node remains with parent
            #ifdef VERBOSE
            std::cout << "Child " << i << " remains with existing parent with distance " << dist_to_a << " to a (versus " << dist_to_b << " to b)" << std::endl;
            #endif
        }
    }


    //distribute newnode
    dist_to_a = dist_between_nodes(newnode, a);
    dist_to_b = dist_between_nodes(newnode, b);

    if (dist_to_b < dist_to_a){ //newnode has to be appended to newparent
        newparent->append_child(newnode); //not necessary to call remove_child because newnode hasn't been appended to any parent node before
        #ifdef VERBOSE
        std::cout << "Appending newly created node to new parent with distance " << dist_to_b << " to b (versus " << dist_to_a << " to a)" << std::endl;
        #endif
    }
    else{ //newnode has to be appended to parent
        parent->append_child(newnode);
        #ifdef VERBOSE
        std::cout << "Newly created node remains with existing parent with distance " << dist_to_a << " to a (versus " << dist_to_b << " to b)" << std::endl;
        #endif
    }
}




//public method that calls the private ones
void CFTree::insert(const float_type *data, int num_points){

    this->num_points = num_points; //storing the number of data points for later

    int point_offset = wavelet_length * wavelets_per_point;
    const float_type *point = nullptr;
    Node *leaf = nullptr;

    for (int i = 0; i < num_points; ++i){

        #ifdef VERBOSE
        std::cout << "\n********** Inserting point " << i << " **********" << std::endl;
        #endif

        point = data + i*point_offset;
        leaf = find_leaf(point);
        modify_leaf(point, leaf, i);

        #ifdef VERBOSE
        print();
        #endif
    }
}




//-------------------------------------getter for clustering results----------------------------------------------------------------


//allocates the memory for cluster_assignments and centroids -> memory has to be freed by the caller
bool CFTree::get_clustering_results(int *k, int *&cluster_assignments, float_type *&centroids, int *&points_per_cluster){
    if (num_points == 0) return false; //no sense in returning clustering results if no clustering has been performed

    //get the resulting number of clusters (=k parameter for I-K-means)
    int num_clusters = leaves.size(); //number of leaf nodes = number of clusters
    *k = num_clusters;

    //memory allocation
    cluster_assignments = new int[num_points];
    centroids = new float_type[num_clusters * wavelet_length * wavelets_per_point](); //allocating memory for the FULL wavelet length, not just what was used for creating the tree => can be used with I-K-means
    points_per_cluster = new int[num_clusters];

    int cluster_id = 0; //cluster IDs depend on the order in which they were added to leaves vector
    for (Node *leaf: leaves){

        //get cluster assignments
        for (int point_id: leaf->assigned_points){
            cluster_assignments[point_id] = cluster_id;
        }
        points_per_cluster[cluster_id] = leaf->n;

        //get the point's centroid
        leaf->calculate_centroid();
        for (int i = 0; i < wavelets_per_point; ++i){
            for (int j = 0; j < used_resolution; ++j){
                centroids[cluster_id*wavelet_length*wavelets_per_point + i*wavelet_length + j] = leaf->centroid[i*used_resolution + j];
            }
        }

        cluster_id++;
    }

    return true;
}







//----------------------------------------------------utility methods---------------------------------------------------------


//add the coordinates of the given point to the linear sum of the given cluster
void CFTree::add_to_linsum(Node *node, const float_type *point){
    int point_data_length = wavelet_length * wavelets_per_point;
    int nodeindex = 0;
    for (int i = 0; i < point_data_length; i += wavelet_length){
        for (int j = 0; j < used_resolution; j++){
            node->linsum[nodeindex++] += point[i+j];
        }
    }
}

//set the coordinates of the given point as the linear sum of the given cluster
void CFTree::set_as_linsum(Node *node, const float_type *point){
    int point_data_length = wavelet_length * wavelets_per_point;
    int nodeindex = 0;
    for (int i = 0; i < point_data_length; i += wavelet_length){
        for (int j = 0; j < used_resolution; j++){
            node->linsum[nodeindex++] = point[i+j];
        }
    }
}

//calculate the sum of the squared coordinates of the given point
float_type CFTree::calculate_squaredsum(const float_type *point){
    float_type squaredsum = 0;
    int point_data_length = wavelet_length * wavelets_per_point;
    for (int i = 0; i < point_data_length; i += wavelet_length){
        for (int j = 0; j < used_resolution; j++){
            squaredsum += pow(point[i+j], 2); //sum of squares
        }
    }
    return squaredsum;
}

//compute the combined diameter of the given cluster and point (i.e. the diameter the cluster would have after the point has been inserted)
float_type CFTree::compute_diameter(Node *node, const float_type *point){
    float_type squaredsum_point = calculate_squaredsum(point);
    int n1 = node->n;
    //n2 = 1, because the second "cluster" consists only of the given point

    //calculate LS1 * LS2
    float_type dotproduct = 0;
    int point_data_length = wavelet_length * wavelets_per_point;
    int nodeindex = 0;
    for (int i = 0; i < point_data_length; i += wavelet_length){
        for (int j = 0; j < used_resolution; j++){
            dotproduct += node->linsum[nodeindex++] * point[i+j];
        }
    }

    //( N1*SS2 + N2*SS1 - 2*LS1*LS2) / (N1 * N2)
    float_type diameter = ( n1 * squaredsum_point + node->squaredsum - 2 * dotproduct ) / n1;
    return diameter;
}

//distance between the centroids of two clusters (= sum of squared coordinate-wise differences)
float_type CFTree::dist_between_nodes(Node *node1, Node *node2){
    //assumption: calculate_centroid() has been called for both nodes
    const float_type *centroid1 = node1->get_centroid();
    const float_type *centroid2 = node2->get_centroid();

    int dim = node1->dim;
    float_type dist = 0;
    float_type help;
    for (int i = 0; i < dim; ++i){
        help = centroid1[i] - centroid2[i];
        dist += help*help;
    }
    return dist;
}

//distance between a point and the centroid of a cluster (= sum of squared coordinate-wise differences)
float_type CFTree::dist_to_centroid(const float_type *point, const float_type *centroid){
    int point_data_length = wavelet_length * wavelets_per_point;
    int cindex = 0;
    float_type dist = 0;
    float_type help;
    for (int i = 0; i < point_data_length; i += wavelet_length){
        for (int j = 0; j < used_resolution; j++){
            help = point[i+j] - centroid[cindex++];
            dist += help*help;
        }
    }
    return dist;
}


//
//printing
//

void CFTree::print(){
    std::cout << '\n';
    if (root){
        root->print("", true);
    }
    else{
        std::cout << "CFTree is empty";
    }
    std::cout << std::endl;
}

void CFTree::print_leaves(){
    for (Node *leaf: leaves){
        leaf->print_self();
    }
}
