#include <iostream>
#include <fstream>
#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>
#include <pybind11/stl.h>
#include <string>
#include <chrono>

#include "inc_cluster.h"
#include "cftree.h"
#include "ikmeans.h"

namespace py = pybind11;

int Incremental_clustering::cluster(py::array_t<float_type, py::array::f_style | py::array::forcecast> wavelet_data, //require data to be laid out contiguously in memory, in column-major-order
    std::vector<std::string> point_ids,
    py::array_t<int, py::array::f_style | py::array::forcecast> coeff_lengths_data,
    int num_points, int wavelets_per_point, int wavelet_length, float_type threshold, int branchingfactor,
    int num_levels /*number of resolution levels in the wavelets*/, int start_with_levels, /*number of levels that should be used for building the CFTree*/
    int max_iterations,
    std::string tag //tag to be appended to the output file name
){

    //begin of time measurement
    std::chrono::steady_clock::time_point begin = std::chrono::steady_clock::now();


    //getting pointers to the python arrays
    float_type *wavelets = static_cast<float_type *>(wavelet_data.request().ptr); //the actual data to be clustered
    int *coeffs_lengths = static_cast<int *>(coeff_lengths_data.request().ptr); //array containing information about the shape of the wavelets


    //-------------------------------------------parameter checks---------------------------------------------------------------

    if (threshold <= 0){
        std::cerr << "threshold must be greater than 0" << std::endl;
        exit(1);
    }
    if (branchingfactor < 2){
        std::cerr << "branchingfactor must be at least 2" << std::endl;
        exit(1);
    }
    if (start_with_levels <= 0){
        std::cerr << "start_with_levels must be greater than 0" << std::endl;
        exit(1);
    }
    if (num_levels <= start_with_levels){
        std::cerr << "start_with_levels must be less than num_levels" << std::endl;
        exit(1);
    }
    if (max_iterations <= 0){
        std::cerr << "max_iterations must be greater than 0" << std::endl;
        exit(1);
    }


    //-------------------------------------------build clustering feature tree--------------------------------------------------


    //get starting resolution: sum up the number of elements in all used levels

    int current_resolution = 0;
    for (int i = 0; i < start_with_levels; ++i){
        current_resolution += coeffs_lengths[i];
    }


    //build the tree

    CFTree tree(threshold, wavelet_length, current_resolution, wavelets_per_point, branchingfactor);
    tree.insert(wavelets, num_points);


    //get clustering results

    int k = 0;
    int *cluster_assignments = nullptr;
    float_type *centroids = nullptr;
    int *points_per_cluster = nullptr;

    bool fetched_results = tree.get_clustering_results(&k, cluster_assignments, centroids, points_per_cluster);
    if (!fetched_results){
        std::cerr << "Could not fetch results" << std::endl;
        exit(1);
    }






    //----------------------------------------perform I-K-means clustering---------------------------------------------------


    IKmeans ikmeans(k, max_iterations, wavelets, num_points, wavelets_per_point, wavelet_length,
    cluster_assignments, centroids, points_per_cluster, coeffs_lengths, num_levels, start_with_levels);
    int iterations = ikmeans.cluster();

    //nd of time measurements (to exclude file I/O)
    std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();

    //print some cluster result details
    std::cout << "Iterations: " << iterations << std::endl;

    //print clustering time
    std::cout << "Time needed for clustering: " << (std::chrono::duration_cast<std::chrono::microseconds>(end - begin).count()) /1000000.0 << "seconds" << std::endl;


    //------------------------------------------------output results-------------------------------------------------------------

    //compute total number of dimensions in the data
    int dim = 0;
    for (int i = 0; i < num_levels; ++i) dim += coeffs_lengths[i];

    bool success;
    if (iterations <= max_iterations)
        success = write_data_to_files(num_points, point_ids, cluster_assignments, threshold, start_with_levels, num_levels, k, iterations, true, tag, dim);
    else
        success = write_data_to_files(num_points, point_ids, cluster_assignments, threshold, start_with_levels, num_levels, k, max_iterations, false, tag, dim);

    if (success){ //data has been written successfully to the output files
        std::cout << "Clustering complete!" << std::endl;
    }



    //----------------------------------------------free memory-----------------------------------------------------

    delete[] cluster_assignments;
    delete[] centroids;
    delete[] points_per_cluster;

    //return the number of clusters
    return k;
}




bool Incremental_clustering::write_data_to_files(int num_points, std::vector<std::string> point_ids, int *cluster_assignments,
float_type threshold, int start_with_levels, int num_levels, int k, int iterations, bool converged, std::string tag, int dim){
    std::string datafilename = tag + "cluster_assignments.csv";
    std::string metafilename = tag + "clustering_metadata.json";

    std::ofstream datafile(datafilename);
    datafile << "STATNR,CLUSTERID\n";
    for (int i = 0; i < num_points; ++i)
        datafile << point_ids[i] << "," << cluster_assignments[i] << '\n';
    datafile.close();

    //printing the metadata in JSON format
    std::ofstream metafile(metafilename);
    metafile << "{\n\"settings\":{\n"
        << "\"number of data points\": " << num_points << ",\n"
        << "\"threshold\": " << threshold << ",\n"
        << "\"CFTree wavelet levels\": " << start_with_levels << ",\n"
        << "\"available wavelet levels\": " << num_levels << "\n"
        << "},\n"
        << "\"results\":{\n"
        << "\"number of clusters\": " << k << ",\n"
        << "\"iterations\": " << iterations << ",\n"
        ;
    if (converged){
        metafile << "\"converged\": true\n";
    }
    else {
        metafile << "\"converged\": false\n";
    }
    metafile << "}\n}";

    metafile.close();

    if (!datafile || !metafile){
        std::cerr << "Could not create output files!" << std::endl;
        return false;
    }
    return true;
}
